/* ref https://github.com/readthedocs/sphinx_rtd_theme/issues/117#issuecomment-153083280
*/
/* https://stackoverflow.com/questions/23211695/modifying-content-width-of-the-sphinx-theme-read-the-docs 

and https://github.com/readthedocs/sphinx_rtd_theme/issues/295

*/
/* override table width restrictions */
@media screen and (min-width: 767px) {
  .wy-nav-content {
      max-width: 1500px !important;
  }


  /* and fix wrap bug per https://rackerlabs.github.io/docs-rackspace/tools/rtd-tables.html */

  .wy-table-responsive table td {
    /* !important prevents the common CSS stylesheets from
       overriding this as on RTD they are loaded after this stylesheet */
    white-space: normal !important;
  }

  .wy-table-responsive {
    overflow: visible !important;
  }

}
/*
.wy-nav-content {
    max-width: 1100px !important;
}
*/
