# -*- coding: utf-8 -*-

WebAPI = ('backup',
          'dereference',
          'exists',
          'getAllUrns', 'getCacheInfo', 'getCount', 'getDefinition', 'getHead',
          'getId', 'getLastVersion', 'getPlace',
          'getPoolname', 'getPoolurl', 'getPoolpath',
          'getProductClasses', 'getReferenceCount',
          'getScheme', 'getTagUrnMap', 'getTags',
          'getUrn', 'getUrnId', 'getUrnObject', 'getVersions',
          'isAlive', 'isEmpty',
          'jsonPath',
          'loadDescriptors', 'loadProduct', 'lockpath',
          'meta', 'mfilter',
          'pfilter', 'poolname', 'poolurl',
          'qm',
          'readHK', 'reference',
          'remove', 'removeAll', 'removeTag', 'removeUrn', 'removekey',
          'restore',
          'saveDescriptors', 'saveProduct', 'saveProductRef',
          'schematicLoad', 'schematicRemove', 'schematicSave', 'schematicSelect', 'schematicWipe',
          'select', 'setTag', 'setup',
          'tagExists',
          'where', 'writeHK')
