# -*- coding: utf-8 -*-

#from fdi.pal.poolmanager import PoolManager
#from fdi.pns.jsonio import getJsonObj, postJsonObj, putJsonObj, commonheaders
#from fdi.utils import getconfig
#from fdi.utils.common import lls
#
#import pytest
#import os
#import logging
##logger = logging.getLogger(__name__)
#
